# python3-opsi
# This code is part of the opsi.org project
# http://opsi.org
#
# Copyright (c) uib gmbh (www.uib.de)
# This sourcecode is owned by the uib gmbh, D55118 Mainz, Germany
#
# This source code is part of co-funding project, and so are
# ALL RIGHTS RESEVERD right now
# http://uib.de/en/opsi_cofunding
# Once the co-funding is finished it will be released as full open source

try:
    import ConfigParser
except:
    import configparser as ConfigParser
from OPSI.Util.File import IniFile
from OPSI.Backend.BackendManager import BackendManager
from OPSI.Backend.JSONRPC import JSONRPCBackend
from OPSI.Util import md5sum
from hashlib import md5
try:
    from twisted.conch.ssh import keys
except:
    pass

if not bootimageVersion or (int(bootimageVersion) < 20160517):
    raise Exception(u"This product requires opsi-linux-bootimage version >= 20160517")
try:
    if not backend.backend_info()['opsiVersion'].startswith('4.'):
        raise ValueError('No opsi 4.x backend present')
except Exception:
    raise Exception(u"This is a opsi 4.x product")

isOpsi43 = False
if backend.backend_info()['opsiVersion'].startswith('4.3'):
    isOpsi43 = True

#include opsi setup library
depot.copy(SCRIPT_PATH+'/opsisetuplib.py','/tmp')
exec(open('/tmp/opsisetuplib.py').read())

checkModules('local_imaging')

logger.notice(u"##############################################################")
initOpsiSetupLib()
logger.notice(u"opsi-local-image-nt6.py")
logger.notice(u"##############################################################")
productVersion =''
packageVersion =''
for i in range(len(productOnDepots)):
    if (productOnDepots[i]['productId'] == productId):
        productVersion = productOnDepots[i]['productVersion']
        packageVersion = productOnDepots[i]['packageVersion']
    break
if isOpsi43:
    show_message("Installing Product: %s_%s-%s on %s" % (productId,productVersion,packageVersion,clientId), 10)
logger.notice(u"Installing Product: %s_%s-%s on %s" % (productId,productVersion,packageVersion,clientId))
scriptMessageSubject.setMessage(u"Installing Product: %s_%s-%s on %s" % (productId,productVersion,packageVersion,clientId))
logger.notice(u"##############################################################")

logger.notice(u"Using built in partclone")


def saveImage(partition, imageFile, progressSubject=None, raw=False, diskDevice = "", fileSystem = "", useGpt = False):
    if not useGpt:
        for hook in hooks:
            (partition, imageFile, progressSubject) = hook.pre_Harddisk_saveImage(disk, partition, imageFile, progressSubject)

    try:
        partition = forceInt(partition)
        imageFile = forceUnicode(imageFile)

        imageType = None
        image = None

        saveImageResult = {'TotalTime': 'n/a','AveRate': 'n/a', 'AveUnit': 'n/a',}

        if not useGpt:
            part = disk.getPartition(partition)
            if not part:
                if isOpsi43:
                    show_error("Partition %s does not exist" % partition)
                raise Exception(u'Partition %s does not exist' % partition)

        if disk.ldPreload:
            os.putenv("LD_PRELOAD", disk.ldPreload)

        pipe = u''
        if imageFile.startswith(u'|'):
            pipe = imageFile
            imageFile = u'-'

        clonemethod = u'--clone'
        if raw:
            clonemethod = u'--dev-to-dev --restore_raw_file'
            realimagefile = imageFile
        else:
            realimagefile = '%s.img' % imageFile
            #if fileSystem == 'ext4':
            #       clonemethod = u'--dev-to-dev --restore_raw_file'

        if useGpt:
            logger.info( u"Saving partition '%s' to partclone image '%s'" % (getPartitionDevicePath(disk.device, partition), realimagefile) )
        else:
            logger.info( u"Saving partition '%s' to partclone image '%s'" % (part['device'], realimagefile) )


        # "-f" will write images of "dirty" volumes too
        # Better run chkdsk under windows before saving image!
        # use here imagefile because partclone will add .img extension
        if useGpt:
            cmd = u'%s --rescue %s --force --source %s --overwrite %s %s' % (which('partclone.' + fileSystem), clonemethod, getPartitionDevicePath(disk.device, partition) , realimagefile, pipe)
        else:
            cmd = u'%s --rescue %s --force --source %s --overwrite %s %s' % (which('partclone.' + part['fs']), clonemethod, part['device'], realimagefile, pipe)

        if progressSubject:
            progressSubject.setEnd(100)

        handle = execute(cmd, getHandle = True)
        done = False

        timeout = 0
        buf = [u'']
        lastMsg = u''
        started = False
        while not done:
            inp = handle.read(128)

            if inp:
                inp = inp.decode("latin-1")
                timeout = 0

                b = inp.splitlines()
                if inp.endswith(u'\n') or inp.endswith(u'\r'):
                    b.append(u'')

                buf = [ buf[-1] + b[0] ] + b[1:]

                for i in range( len(buf)-1 ):
                    try:
                        logger.debug(u" -->>> %s" % buf[i])
                    except Exception:
                        pass
                    if ( buf[i].find(u'Partclone fail') != -1 ):
                        raise Exception(u"Failed: %s" % '\n'.join(buf))
                    if ( buf[i].find(u'Partclone successfully') != -1 ):
                        done = True
                    if ( buf[i].find(u'Total Time') != -1 ):
                        match =  re.search('Total\sTime:\s(\d+:\d+:\d+),\sAve.\sRate:\s*(\d*.\d*)([GgMm]B/min)', buf[i])
                        if match:
                            rate = match.group(2)
                            unit = match.group(3)
                            if unit.startswith("G") or unit.startswith("g"):
                                rate = float(rate.replace(',', '.')) * 1024
                                unit = 'MB/min'
                            saveImageResult = { 'TotalTime': match.group(1),'AveRate': str(rate), 'AveUnit': unit, }
                    elif 'Device size' in buf[i]:
                        match = re.match('Device size:\s+(?P<size>\d+.{0,1}\d+) (?P<unit>\w+)', buf[i])
                        if match:
                            size = float(match.group('size').replace(',', '.'))
                            unit = match.group('unit')
                            saveImageResult['DeviceSize'] = size
                            saveImageResult['DeviceSizeUnit'] = unit
                    elif 'Space in use' in buf[i]:
                        match = re.match('Space in use:\s+(?P<size>\d+.{0,1}\d+) (?P<unit>\w+)', buf[i])
                        if match:
                            size = float(match.group('size').replace(',', '.'))
                            unit = match.group('unit')
                            saveImageResult['SpaceInUse'] = size
                            saveImageResult['SpaceInUseUnit'] = unit
                    if not started:
                        if ( buf[i].find(u'Calculating bitmap') != -1 ):
                            logger.info(u"Save image: Scanning filesystem")
                            if progressSubject:
                                progressSubject.setMessage(u"Scanning filesystem")
                        elif ( buf[i].count(':') == 1 ) and (buf[i].find('http:') == -1):
                            (k, v) = buf[i].split(':')
                            k = k.strip()
                            v = v.strip()
                            logger.info(u"Save image: %s: %s" % (k, v))
                            if progressSubject:
                                progressSubject.setMessage(u"%s: %s" % (k, v))
                            if (k.lower().find('used') != -1) or (k.lower().find('space in use') != -1):
                                if progressSubject:
                                    progressSubject.setMessage(u"Creating image")
                                started = True
                                continue
                    else:
                        match = re.search('Completed:\s*([\d\.,]+)%', buf[i])
                        if match:
                            percent = int("%0.f" % float(match.group(1).replace(',','.')))
                            if isOpsi43:
                                show_message("backup running", percent)
                            if progressSubject and (percent != progressSubject.getState()):
                                logger.debug(u" -->>> %s" % buf[i])
                                progressSubject.setState(percent)
                try:
                    # sometimes index out of range here
                    lastMsg = buf[-2]
                except:
                    lastMsg = u''

                buf[:-1] = []

            elif (timeout >= 100):
                if isOpsi43:
                    show_error("Failed: %s" % lastMsg)
                raise Exception(u"Failed: %s" % lastMsg)
            else:
                timeout += 1
                continue

        time.sleep(3)
        if handle: handle.close()

        if disk.ldPreload:
            os.unsetenv("LD_PRELOAD")
    except Exception as e:
        if not useGpt:
            for hook in hooks:
                hook.error_Harddisk_saveImage(disk, partition, imageFile, progressSubject, e)
        raise

    if not useGpt:
        for hook in hooks:
            hook.post_Harddisk_saveImage(disk, partition, imageFile, progressSubject)

    return saveImageResult

useGpt = False
inUefi = inUefiMode()
if inUefi:
    useGpt = True

setupAfterInstall = productPropertyValues.get('setup_after_install', '')

partgapM = 1024
systemPartitionNumber = 1
winpePartitionNumber = 2
swapPartitionLabel  = "WINPE"
backupPartitionNumber = 3
backupPartitionLabel  = "opsidata"
if useGpt:
    systemPartitionNumber += 2
    winpePartitionNumber += 2
    backupPartitionNumber += 2

method = 'partclone-image'
dataPartitionSizeM = 0
dataPartitionNumber = 0
dataPartitionLabel  = 'sysdata'

blockAlignment = True
useRaid1 = False
goOn = True

logger.notice("Partition numbers are now: system=%d, data=%s, swap=%d, backup=%d" % (systemPartitionNumber, dataPartitionNumber, winpePartitionNumber, backupPartitionNumber))

# get multiDiskMode and backup_partition_on_same_disk from opsi-local-image-prepare
multiDiskMode  = '0'
backup_partition_on_same_disk = True
if hasattr(backend, "productPropertyState_getValues"):
    # opsi >= 4.3
    for property_id, values in (
        backend.productPropertyState_getValues(product_ids=['opsi-local-image-prepare'], object_ids=[clientId], with_defaults=True)[clientId]['opsi-local-image-prepare']
        .items()
    ):
        logger.info("Found product property: %s=%r", property_id, values)
        if property_id == 'multi_disk_mode':
            multiDiskMode  = values[0]
        if property_id == 'backup_partition_on_same_disk':
            backup_partition_on_same_disk = values[0]
else:
    backend.backend_setOptions( { 'addProductPropertyStateDefaults': True } )
    pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'multi_disk_mode')
    if pps:
        for state in pps:
            myvalues = state.getValues()
            for key in myvalues:
                multiDiskMode  = key
    pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-prepare',propertyId = 'backup_partition_on_same_disk')
    if pps:
        for state in pps:
            myvalues = state.getValues()
            for key in myvalues:
                backup_partition_on_same_disk  = key
    backend.backend_setOptions( { 'addProductPropertyStateDefaults': False } )
    
logger.notice("multiDiskMode: %s ; backup_partition_on_same_disk: %s" % (multiDiskMode,backup_partition_on_same_disk))

if isOpsi43:
    show_message("Reading disks", 15)

logger.notice(u"Swapoff all to avoid disk in use .....")
execute("swapoff -a -v")
#lvmClearVolumeGroups()
# Get harddisks
disks = getHarddisks()
disks = sorted(disks, key=lambda disk: disk.device)
showDiskOrder(disks)
logger.notice(u"Normally we using disk 0")
diskindex = getDiskIndexFromMultiDiskMode(multiDiskMode)
disk = disks[diskindex]
logger.notice(u"Using disk device %s" % (disk.device))
# check if (backup_partition_on_same_disk == False) and only one disk, what would make no sense
# and if so fall back to (backup_partition_on_same_disk == true)
backupdiskindex, backup_partition_on_same_disk = getBackupdiskindexFromBackupPartitionOnSameDisk(backup_partition_on_same_disk)
logger.notice(u"backup_partition_on_same_disk: %s - backupdiskindex %s" % (backup_partition_on_same_disk,backupdiskindex))
if (backup_partition_on_same_disk == False):
    # so we have a backup disk 
    backupPartitionNumber = 1

logger.notice("Partition numbers are now: system=%d, data=%s, swap=%d, backup=%d" % (systemPartitionNumber, dataPartitionNumber, winpePartitionNumber, backupPartitionNumber))

if not useGpt:
    # Show partition information
    partitions = disk.getPartitions()
    for p in partitions:
        fs   = p.get('fs', 'unknown')
        mytype = p.get('type', 'unknown')
        scriptMessageSubject.setMessage( u"Partition %s: Start %s MB, End %s MB, Filesystem %s, Typ %s.\n" \
                                % (     p['number'],
                                        p['start']/(1000*1000),
                                        p['end']/(1000*1000),
                                        fs,
                                        mytype) )
else:
    execute("sgdisk -p %s" % disk.device)
    execute("/bin/efibootmgr -v")
    uefiShowEntriesByBootOrder()
    firstbootentry = uefiGetFirstBootEntry()
    if firstbootentry:
        logger.notice('First boot entry: {0}'.format(firstbootentry['label']))
        if 'HD' in firstbootentry['parameters']:
            logger.notice('Partition: {0}'.format(firstbootentry['parameters']['HD'][0]))
            logger.notice('Loader: {0}'.format(firstbootentry['parameters']['File']))
    else:
        logger.notice('No usable first boot entry found')
    currentbootentry = uefiGetActBootEntry()
    if currentbootentry:
        logger.notice('Current boot entry: {0}'.format(currentbootentry['label']))
        if 'device' in currentbootentry['parameters']:
            logger.notice('device: {0}'.format(currentbootentry['parameters']['device']))
    else:
        logger.notice('No usable current boot entry found')


# Get partition number to backup
partition = 1
if useGpt:
    partition += 2
    partitionLabel = partitionHasLabel(disk.device, 3, getPartitionDevicePath(disk.device, 3))

logger.notice("We will backup partition: %d from disk %d to partition %d on disk %d" % (partition,diskindex,backupPartitionNumber,backupdiskindex))


imageDir = '/mntimg'
prodDir = '/mntprod'
product = 'unknown'
imageFileOrg = productProperties.get('imagefile', '')
imageFile =forceUnicode(re.sub("[\s\<\>\?\"\:\|\\\/\*]", "_",imageFileOrg))
logger.notice(u"Original image file name >%s< changed to internal image file name >%s<" % (imageFileOrg, imageFile))
logger.notice(u"getting name of product to backup")
for poc in backend.productOnClient_getObjects(clientId = clientId, productType = 'NetbootProduct'):
    if (poc.installationStatus == 'installed'):
        if not (poc.productId in ['opsi-local-image-backup', 'opsi-local-image-restore', 'hwinvent', 'opsi-local-image-delimage']):
            product = poc.productId
logger.notice(u"name of product: %s" % product)
#imageFile = ''
if (imageFile == ''):
    logger.notice(u"no image name given - using name of product to backup")
    imageFile = product
    imageFilePublish = imageFile
    goOn = True
else:
    # we should publish the not patched filename
    imageFilePublish = imageFileOrg


logger.notice(u"will backup to image file name %s" % imageFile)
scriptMessageSubject.setMessage(u"will backup to image file name %s " % imageFilePublish)

# detect filesystem of partition to backup
filesystem = 'unknown'
tempresultlist = execute(u'blkid -o value -s TYPE %s' % (getPartitionDevicePath(disk.device, partition)))
resultlist = filter(None, tempresultlist)
for resultline in resultlist:
    try:
        filesystem = resultline.strip()
    except:
        pass
logger.notice(u"%s exists as device with filesystem: %s" % (getPartitionDevicePath(disk.device, partition), filesystem) )


# mount the image (backup) file system
# check the mount point
exitcode = os.system("test -d %s >> /dev/null 2>&1" % (imageDir))
if (0 == int(exitcode)):
    logger.notice(u"%s exists as directory - fine" % imageDir)
else:
    logger.notice(u"%s does not exists as directory - creating" % imageDir)
    os.makedirs(imageDir)
    #os.system("mkdir %s" % (imageDir))

checkOliPartitions(diskindex,backupdiskindex,useGpt,backup_partition_on_same_disk)
disk = disks[backupdiskindex]

# check if mount point is in use
checkCreateDir(imageDir)

try:
    testPartitionEx(disk.device, backupPartitionNumber, useGpt)
except:
    raise Exception(u"Could not find partition %d. - Please restart this product (opsi-local-image-prepare)" % backupPartitionNumber)

partitionLabel = partitionHasLabel(disk.device, backupPartitionNumber,getPartitionDevicePath(disk.device, backupPartitionNumber))
if (partitionLabel == backupPartitionLabel):
    logger.notice(u"%s exists as expected - fine" % (getPartitionDevicePath(disk.device, backupPartitionNumber)) )
    logger.notice(u"%s created and formated" % (getPartitionDevicePath(disk.device, backupPartitionNumber)))
else:
    logger.error(u"%s exists not as expected with label %s - not ok - giving up" % (getPartitionDevicePath(disk.device, backupPartitionNumber), backupPartitionLabel) )
    scriptMessageSubject.setMessage(u"%s exists not as expected with label %s - not ok - giving up" % (getPartitionDevicePath(disk.device, backupPartitionNumber), backupPartitionLabel) )
    raise Exception(u"Wrong partition table - Please run the opsi product 'opsi-local-image-prepare' to prepare this disk ! ")

# check if backup partition is in use
checkClearMountpoint(imageDir, disk.device, backupPartitionNumber)
checkClearPartition(imageDir, disk.device, backupPartitionNumber)
# mount backup partition to mount point
checkMount(imageDir, disk.device, backupPartitionNumber)

logger.notice(u"disk free on %s" % (imageDir))
execute("df -h %s" % imageDir)

logger.notice("switch back to system disk: %d" % (diskindex))
disk = disks[diskindex]

if goOn:
    # check the backup directory (which has the name of the imageFile)
    exitcode = os.system("test -d %s/%s >> /dev/null 2>&1" % (imageDir, imageFile))
    if (0 == int(exitcode)):
        logger.notice(u"%s/%s exists as directory - fine" % (imageDir, imageFile))
    else:
        logger.notice(u"%s/%s does not exists as directory - creating" % (imageDir, imageFile))
        os.makedirs(u"%s/%s" % (imageDir, imageFile))
        #os.system("mkdir %s" % (imageDir))


if goOn:
    def _getProgressSubject():
        from OPSI.Util.Message import ProgressSubject
        return ProgressSubject(id=u'backup_creation', title=u'Creating Backup', end=100, fireAlways=True)

    try:
        if progressSubject is None:
            progressSubject = _getProgressSubject()
    except NameError:
        # No object defined.
        progressSubject = _getProgressSubject()

    backupProgressBox = ui.createProgressBox(width=60, height=20, total=100, title='Creating Backup', text=u'')
    logger.debug('Created Progress box: {0}'.format(backupProgressBox))
    progressSubject.attachObserver(backupProgressBox)

    if (method == "partclone-image"):
        if (filesystem == 'ntfs')or (filesystem == 'ext4'):
            logger.notice(u"creating a partclone image from %s to %s/%s/%s" % (getPartitionDevicePath(disk.device, partition), imageDir, imageFile, imageFile) )
            if os.path.exists("%s/%s/%s" % (imageDir,imageFile,imageFile)):
                myoldimagefile = "%s/%s/%s" % (imageDir,imageFile,imageFile)
                try:
                    execute("rm %s" % (myoldimagefile))
                    logger.notice('Old image file %s removed' % (myoldimagefile))
                except:
                    pass
            if os.path.exists("%s/%s/%s.img" % (imageDir,imageFile,imageFile)):
                myoldimagefile = "%s/%s/%s.img" % (imageDir,imageFile,imageFile)
                try:
                    execute("rm %s" % (myoldimagefile))
                    logger.notice('Old image file %s removed' % (myoldimagefile))
                except:
                    pass
            #execute("partclone.ntfs -C -b -s %s%s -O %s/%s/%s" % (disk.device, partition, imageDir, imageFile, imageFile) )
            #saveImage(partition, imageFile, progressSubject=None, raw=False, diskDevice = "", fileSystem = "", useGpt = False):
            #resultlist = saveImage(partition = int(partition), raw=True, imageFile = u"%s/%s/%s" % (imageDir, imageFile, imageFile), diskDevice = disk.device, fileSystem = filesystem, useGpt = useGpt, progressSubject=progressSubject)
            resultlist = saveImage(partition = int(partition), raw=False, imageFile = u"%s/%s/%s" % (imageDir, imageFile, imageFile), diskDevice = disk.device, fileSystem = filesystem, useGpt = useGpt, progressSubject=progressSubject)
            #resultlist = disk.saveImage( partition = int(partition), imageFile = u"%s/%s/%s" % (imageDir, imageFile, imageFile) )

            AveRate = resultlist.get('AveRate', '')
            AveUnit = resultlist.get('AveUnit', '')
            TotalTime = resultlist.get('TotalTime', '')
            logger.notice(u"AveRate: %s " % (AveRate) )
            logger.notice(u"AveUnit: %s " % (AveUnit) )
            logger.notice(u"TotalTime: %s " % (TotalTime) )


    progressSubject.detachObserver(backupProgressBox)
    backupProgressBox.hide()

if goOn:
    # write meta info file
    logger.notice(u"Setting defaults of restore to image: %s " % (imageFile) )
    iniFile = imageDir+'/'+imageFile+'/img.ini'
    iniFileHandler = IniFile(iniFile)
    if not os.path.exists(iniFile):
        # Wenn Datei nicht existiert generiere leeren ConfigParser
        config = ConfigParser.RawConfigParser()
        # initiales fuellen der iniDatei und wegschreiben.
    else:
        # Wenn Datei existiert parse Datei und gebe ConfigParser Objekt zurueck
        config = iniFileHandler.parse()

    # create section if not there
    if not config.has_section("%s" % imageFile):
        config.add_section("%s" % imageFile)
    # write data
    logger.notice(u"Setting image meta data: given imgage file name: %s " % (imageFileOrg) )
    config.set(imageFile,"imageFileOrg",imageFileOrg)
    logger.notice(u"Setting image meta data: filesystem: %s " % (filesystem) )
    config.set(imageFile,"filesystem",filesystem)
    logger.notice(u"Setting image meta data: opsiproduct: %s " % (product) )
    config.set(imageFile,"opsiproduct",product)
    logger.notice(u"Setting image meta data: imagetype: %s " % (method) )
    config.set(imageFile,"imagetype",method)
    logger.notice(u"Setting image meta data: date: %s " % (time.asctime( time.localtime(time.time()) )) )
    config.set(imageFile,"date",time.asctime( time.localtime(time.time()) ))
    logger.notice(u"Setting image meta data: AveRate: %s " % (AveRate) )
    config.set(imageFile,"AveRate",AveRate)
    logger.notice(u"Setting image meta data: TotalTime: %s " % (TotalTime) )
    config.set(imageFile,"TotalTime",TotalTime)
    logger.notice(u"Setting image meta data: GPT: %s " % (useGpt) )
    config.set(imageFile,"GPT",useGpt)
    if useGpt:
        if firstbootentry:
            logger.notice(u"Setting image meta data: uefiFirstBootLabel: %s " % (firstbootentry['label']) )
            config.set(imageFile,"uefiFirstBootLabel",firstbootentry['label'])
            if 'HD' in firstbootentry['parameters']:
                logger.notice(u"Setting image meta data: uefiFirstBootPartition: %s " % (firstbootentry['parameters']['HD'][0]) )
                config.set(imageFile,"uefiFirstBootPartition",firstbootentry['parameters']['HD'][0])
                logger.notice(u"Setting image meta data: uefiFirstBootLoader: %s " % (firstbootentry['parameters']['File']) )
                config.set(imageFile,"uefiFirstBootLoader",firstbootentry['parameters']['File'])
        else:
            logger.notice('No usable first boot entry found')
        if currentbootentry:
            logger.notice(u"Setting image meta data: uefiCurrentBootLabel: %s " % (currentbootentry['label']) )
            config.set(imageFile,"uefiCurrentBootLabel",currentbootentry['label'])
            if 'device' in currentbootentry['parameters']:
                logger.notice(u"Setting image meta data: uefiCurrentBootDevice: %s " % (currentbootentry['parameters']['device']) )
                config.set(imageFile,"uefiCurrentBootDevice",currentbootentry['parameters']['device'])
        else:
            logger.notice('No usable current boot entry found')

    # save
    iniFileHandler.generate(config)

    # Setting defaults for opsi-local-image-restore
    logger.notice(u"Setting defaults for opsi-local-image-restore to image: %s " % (imageFilePublish) )
    imageFileType = type(imageFilePublish)
    # set imagefile
    myvalue = [ imageFilePublish ]
    backend.productPropertyState_create(productId =  'opsi-local-image-restore',propertyId = 'imagefile',objectId = clientId,values = myvalue)

    # set imagefiles_list
    #backend.backend_setOptions( { 'addProductPropertyStateDefaults': True } )
    myvalues = []
    if hasattr(backend, "productPropertyState_getValues"):
        myValues = backend.productPropertyState_getValues(product_ids=['opsi-local-image-restore'], object_ids=[clientId], property_ids=['imagefiles_list'], with_defaults=True)
    else:
        pps = backend.productPropertyState_getObjects(objectId = clientId, productId = 'opsi-local-image-restore',propertyId = 'imagefiles_list')
        if pps:
            for state in pps:
                myvalues = state.getValues()
        else:
            myValuesBefore = []
            myvalues = [ imageFilePublish ]
            myvaluesType = type(myvalues)
            backend.productPropertyState_create(productId =  'opsi-local-image-restore',propertyId = 'imagefiles_list',objectId = clientId,values = myvalues)
    myValuesBefore = myvalues
    for key in myvalues:
        if u',' in key:
            # This is the workaround for a bug where the list of
            # images will become a comma-seperated string instead of
            # a list of strings.
            for key_part in key.split(u','):
                if not (key_part in myvalues):
                    myvalues.append(key_part)
            myvalues.remove(key)
            backend.productPropertyState_create(productId =  'opsi-local-image-restore',propertyId = 'imagefiles_list',objectId = clientId,values = myvalues)
    myvaluesType = type(myvalues)

    if not (imageFilePublish in myvalues):
        myvalues.append(imageFilePublish)
        myvalues = forceList(myvalues)
        logger.notice(u"Adding imagefile: %s to property imagefiles_list" % (imageFilePublish) )
        backend.productPropertyState_create(productId =  'opsi-local-image-restore',propertyId = 'imagefiles_list',objectId = clientId,values = myvalues)
    else:
        logger.notice(u"imagefile: %s is still in imagefiles_list" % (imageFilePublish) )

    logger.notice(u"disk free on %s" % (imageDir))
    dfoutput = execute("df -h %s" % imageDir)
    backend.productPropertyState_create(productId =  'opsi-local-image-backup',propertyId = 'free_on_backup',objectId = clientId,values = dfoutput)


    # write meta info to masterlog
    logstr = time.asctime( time.localtime(time.time()) )
    logstr = logstr + ', product=%s, filesystem=%s, imagefile=%s, method=%s' % (productId, filesystem, imageFile, method)
    logstr = logstr + ', AveRate=%s, TotalTime=%s' % (AveRate, TotalTime)
    logstr = logstr + ', imageFileType=%s, myvaluesType=%s' % (imageFileType, myvaluesType)
    logstr = logstr + ', myValuesBefore=%s, myvalues=%s' % (myValuesBefore, myvalues)
    logstr = logstr + ', free_on_backup=%s' % (dfoutput)
    logFile = imageDir+'/master.log'
    if not os.path.exists(logFile):
        f = open(logFile, 'w' )
    else:
        f = open(logFile, 'a' )
    f.write(logstr)
    f.close()
    # show
    logger.notice(u"Here comes the master log:")
    execute(u'cat %s' % (logFile) )


    # Set product installation status
    productOnClient.setInstallationStatus('installed')
    productOnClient.setActionProgress('')
    productOnClient.setActionResult('successful')
    productOnClient.setLastAction('setup')
    backend.productOnClient_updateObject(productOnClient)

    if (filesystem == 'ntfs') or (filesystem == 'ext4'):
        # write productOnClient
        logger.notice(u"Storing productOnClients data for created image")
        storePocs = []
        # list of all productOnClientds
        for poc in  backend.productOnClient_getObjects(clientId=clientId):
            # remove action requeste
            if (poc.actionRequest != 'none'):
                poc.actionRequest = 'none'
            storePocs.append(poc)
        line = toJson(storePocs)
        pocFile = imageDir+'/'+imageFile+'/productOnClients.json'
        f = open(pocFile, "w")
        f.write(line)
        f.close()

    if setupAfterInstall:
        for setupProduct in setupAfterInstall:
            setupProduct = setupProduct.strip()
            if setupProduct != '':
                logger.notice(u"Set to setup %s" % (setupProduct) )
                backend.setProductActionRequestWithDependencies(productId=setupProduct, clientId=clientId, actionRequest='setup')


else:
    # Set product installation status
    productOnClient.setInstallationStatus('unknown')
    productOnClient.setActionProgress('')
    productOnClient.setActionResult('failed')
    productOnClient.setLastAction('setup')
    backend.productOnClient_updateObject(productOnClient)

# Reboot
reboot()
